from sigla import config
from sigla.nodes.node import Node

from sigla.nodes.node_echo import NodeEcho
from sigla.nodes.node_file import NodeFile
from sigla.nodes.node_template import NodeTemplate

from sigla.templates.engines import JinjaEngine, TemplateEngineABC
from sigla.templates.loaders import FileTemplateLoader, TemplateLoaderABC


class NodeFactory:
    default = NodeTemplate
    reference = {
        "root": Node,
        "file": NodeFile,
        "echo": NodeEcho,
    }

    def __init__(
        self,
        engine: TemplateEngineABC = None,
        loader: TemplateLoaderABC = None,
    ):
        if engine is None:
            engine = JinjaEngine()

        if loader is None:
            loader = FileTemplateLoader(config.path.templates, "jinja2")

        self.engine = engine
        self.loader = loader

    def __call__(self, tag: str, attributes: object):
        node_creator = self.reference.get(tag)

        if node_creator:
            return node_creator(tag, self.engine, attributes=attributes)
        else:
            return self.default(
                tag, self.engine, self.loader, attributes=attributes
            )



from abc import ABC, abstractmethod
from xml.etree import ElementTree
from sigla.utils.type_casters import cast_dict
from sigla.data.data import Data


class DataLoader(ABC):
    @abstractmethod
    def load(self) -> Data:
        pass


class ElementTreeDataLoader(DataLoader):
    def __init__(self, element):
        self.element = element

    def load(self) -> Data:
        raw_attributes = self.element.attrib.copy()

        attributes = cast_dict(raw_attributes)
        children = [
            ElementTreeDataLoader(child).load() for child in self.element
        ]

        return Data(
            tag=self.element.tag, attributes=attributes, children=children
        )


class XMLStringDataLoader(DataLoader):
    def __init__(self, source: str):
        self.source = source

    def load(self) -> Data:
        obj: ElementTree = ElementTree.fromstring(self.source)
        return ElementTreeDataLoader(obj).load()



from textwrap import dedent
from sigla.frontmatter.frontmatter import FrontMatter
from sigla.nodes.node import Node
from sigla.nodes.node_list import NodeList
from sigla.templates.engines import TemplateEngineABC
from sigla.templates.loaders import TemplateLoaderABC
from sigla.utils.errors import TemplateDoesNotExistError


class NodeTemplate(Node):
    create_template = True

    def __init__(
        self,
        tag,
        engine: TemplateEngineABC,
        template_loader: TemplateLoaderABC,
        **kwargs
    ):
        super().__init__(tag, engine, **kwargs)

        self.loader = template_loader

    def finish(self):
        msg = (
            "This is a template node. It should probably not be the first "
            "one to be called"
        )
        raise NotImplementedError(msg)

    def process(self):
        self.data.frontmatter = self._get_frontmatter_from_template()
        super().process()

        # load template for tag
        template = self._get_template_without_frontmatter()

        # handle frontmatter
        return self.render(template)

    #
    #
    #

    def _get_frontmatter_from_template(self):
        raw_template = self._get_raw_template()

        frontmatter = FrontMatter()
        frontmatter_str, template, handler = frontmatter.split(raw_template)

        if not frontmatter or not handler:
            return {}

        content = self.render(frontmatter_str)
        return FrontMatter.parse_with_handler(handler, content)

    def _get_raw_template(self):
        bundle = self.attributes.get("bundle")
        exists = self.loader.exists(self.data.tag, bundle=bundle)
        if not exists:
            if self.create_template:
                default_template = self._create_default_template()
                path = self.loader.get_path(self.data.tag, bundle=bundle)
                path.write_text(default_template)
            else:
                raise TemplateDoesNotExistError(self.data.tag, self)
        return self.loader.read(self.data.tag, bundle=bundle)

    def _get_template_without_frontmatter(self):
        raw_template = self._get_raw_template()
        template = FrontMatter().get_content(raw_template)
        return template

    def _create_default_template(self):

        default_template_content = dedent(
            """\
            ---
            title: Generating code like it's 99
            ---
        
            You can use variables written inside the frontmatter:
                - {% raw %}{{ node.title }}{% endraw %}: Generating code like it's 99
        
            Available variables (with current values for first node as an example):
            {%- for key, value in variables.items() %}
                - {{ "{{ node." + key + " }}" }}: {{value}}
            {%- else %}
                - none for now
            {%- endfor %}
        
            Methods available on node.children:
            {%- for value in methods %}
                - {{ value }}
            {%- endfor %}
        
            You can also iterate over node.children and access variables, render, ...
            {% raw %}
            {% for child in node.children %}
                {{ child() }}
            {% endfor %}
            {% endraw %}
        
            """  # noqa E501
        )

        variables = self.attributes
        # TODO look into this and document it
        methods = NodeList.get_node_list_methods()

        content = self.render(
            default_template_content,
            variables=variables,
            methods=list(methods),
        )
        return content



from abc import ABC, abstractmethod
from typing import Any

from jinja2 import Environment, BaseLoader, Template


class TemplateEngineABC(ABC):
    @abstractmethod
    def render(self, template: str, filters: dict, **kwargs: Any) -> str:
        pass


class JinjaEngine(TemplateEngineABC):
    def render(self, template: str, filters: dict, **kwargs: Any) -> str:
        env = Environment(loader=BaseLoader())
        env.filters.update(filters)
        jinja_tpl: Template = env.from_string(template)
        return jinja_tpl.render(**kwargs)



from abc import ABC, abstractmethod
from pathlib import Path


class TemplateLoaderABC(ABC):
    @abstractmethod
    def get_path(self, tag, *, bundle=None) -> Path:
        pass

    @abstractmethod
    def exists(self, tag, *, bundle=None) -> bool:
        pass

    @abstractmethod
    def read(self, tag, bundle=None) -> str:
        pass


class FileTemplateLoader(TemplateLoaderABC):
    def __init__(self, base_path, ext="jinja2"):
        self.base_path = base_path
        self.ext = ext

    def get_path(self, tag, *, bundle=None) -> Path:
        path = Path(self.base_path)
        if bundle:
            path = path.joinpath(bundle)
        path = path.joinpath(f"{tag}.{self.ext}")
        return path

    def exists(self, tag, *, bundle=None) -> bool:
        path = self.get_path(tag, bundle=bundle)
        return path.exists()

    def read(self, tag, *, bundle=None) -> str:
        path = self.get_path(tag, bundle=bundle)
        return path.read_text()



from abc import abstractmethod, ABC
from sigla.data.loaders import XMLStringDataLoader


class NodeLoader(ABC):
    @abstractmethod
    def load(self):
        pass


class DataToNodeLoader(NodeLoader):
    def __init__(self, data, factory):
        self.data = data
        self.factory = factory

    def load(self):
        ret = self.factory(self.data.tag, self.data.attributes)

        for r in self.data.children:
            loader = DataToNodeLoader(r, self.factory)
            child = loader.load()
            ret.append(child)

        return ret


class XMLToNodeLoader(DataToNodeLoader):
    def __init__(self, data, factory):
        data = XMLStringDataLoader(data).load()
        super().__init__(data, factory)



from sigla.node_loaders import DataToNodeLoader, XMLToNodeLoader
from sigla.nodes.abstract_node import AbstractNode


def process(kind, content, *, factory=None):
    node = load_node(kind, content, factory=factory)
    node.process()
    node.finish()


def load_node(kind, content, *, factory=None) -> AbstractNode:
    if kind == "data":
        return DataToNodeLoader(content, factory).load()
    elif kind == "xml":
        return XMLToNodeLoader(content, factory).load()
    else:
        raise NotImplementedError(f"No loader implemented for {kind}")



from abc import ABC, abstractmethod
import typer
import textwrap
from pathlib import Path


class Action(ABC):
    @abstractmethod
    def run(self):
        pass

    @staticmethod
    def log(msg):
        print(f"|> {msg}")


class NewFileAction(ABC):
    def __init__(self, path, name):
        self.path = path
        self.name = name

    @property
    @abstractmethod
    def content(self) -> str:
        pass

    @property
    def filepath(self):
        return Path(self.path).joinpath(f"{self.name}")

    def run(self):
        if self.filepath.exists():
            raise typer.Exit(f"✋ File {self.filepath} already exists")
        self.filepath.write_text(self.content)


class NewDefinitionFile(NewFileAction):
    extension = "xml"

    def __init__(self, path, name):
        super().__init__(path, name + ".xml")
        self.original_name = name

    @property
    def content(self):
        return textwrap.dedent(
            f"""\
            <root>
                <file to="output/{self.original_name}.txt">
                    <{self.original_name}>
                        [...]
                    </{self.original_name}>
                </file>
            </root>
        """
        )


class NewFiltersFile(NewFileAction):
    @property
    def content(self):
        return textwrap.dedent(
            """
            \"\"\"
            Export filters to use on the templates using the `FILTERS` variable
            \"\"\"
            import json
            from sigla import register_filter

            @register_filter('dump')
            def dump(var):
                return json.dumps(var, indent=4)

            """
        )



import environ


@environ.config(prefix="")
class Config:
    @environ.config
    class Paths:
        root_directory = environ.var(".sigla")
        templates_folder = environ.var("templates")
        snapshots_folder = environ.var("snapshots")
        definitions_folder = environ.var("definitions")
        scripts_folder = environ.var("scripts")

        filters_filename = environ.var("filters.py")

        @property
        def templates(self):
            return f"{self.root_directory}/{self.templates_folder}"

        @property
        def scripts(self):
            return f"{self.root_directory}/{self.scripts_folder}"

        @property
        def snapshots(self):
            return f"{self.root_directory}/{self.snapshots_folder}"

        @property
        def definitions(self):
            return f"{self.root_directory}/{self.definitions_folder}"

        @property
        def filters(self):
            return f"{self.root_directory}/{self.filters_filename}"

    @environ.config
    class Cls:
        node_list = environ.var("sigla.core.node_lists.NodeList")
        node = environ.var("sigla.core.nodes.Node")

    path: Paths = environ.group(Paths)
    cls: Cls = environ.group(Cls)


config = environ.to_config(Config)



import importlib
import types
from contextlib import suppress
from pathlib import Path


def ensure_dirs(*paths: str):
    """Ensure the paths exist"""
    for path_ in paths:
        Path(path_).mkdir(parents=True, exist_ok=True)


def load_module(module_name, module_path):
    loader = importlib.machinery.SourceFileLoader(module_name, module_path)
    module = types.ModuleType(loader.name)
    loader.exec_module(module)
    return module


def load_filters_from(module_path):
    filters = {}
    with suppress(FileNotFoundError):
        filters_module = load_module("filters", module_path)
        if "FILTERS" in dir(filters_module):
            filters = filters_module.FILTERS
    return filters


# import importlib
# import importlib.machinery
# from typing import Type, TypeVar

# T = TypeVar("T")

# def import_reference(ref: str) -> Type[T]:
#     """Imports stuff from modules dynamically"""
#     p, m = ref.rsplit(".", 1)
#     mod = importlib.import_module(p)
#     res: Type[T] = getattr(mod, m)
#     return res



from collections import ChainMap
from copy import deepcopy

from sigla.nodes.node_list import NodeList


class Attributes(ChainMap):
    def without(self, *args):
        data = deepcopy(dict(self))
        for name in args:
            del data[name]
        return Attributes(data)

    def as_kwargs(self, sep=","):
        kwargs = []
        for k, v in self.items():
            if type(v) == int:
                kwargs.append(f"{k}={v}")
            else:
                kwargs.append(f'{k}="{v}"')
        if sep:
            return ", ".join(kwargs)
        return kwargs


class Data(object):
    _meta = [
        {"name": "children"},
        {"name": "tag"},
        {"name": "attributes"},
        {"name": "parent_attributes"},
        {"name": "frontmatter"},
    ]

    def __init__(
        self,
        *,
        children=None,
        tag=None,
        attributes=None,
        parent_attributes=None,
        frontmatter=None,
    ):
        self.tag = tag if tag else {}
        self.children = children if children else NodeList()

        self.attributes = attributes if attributes else {}
        self.parent_attributes = parent_attributes if parent_attributes else {}
        self.frontmatter = frontmatter if frontmatter else {}

    @staticmethod
    def _comp(a, b):
        return (not a and not b) or a == b

    def __eq__(self, other):
        for key in self._meta:
            self_value = getattr(self, key["name"])
            other_value = getattr(other, key["name"])

            if not self._comp(self_value, other_value):
                return False

        return True

    def get_attributes(self):
        return Attributes(
            self.frontmatter, self.attributes, self.parent_attributes
        )



from os import path
from typing import Any
from textwrap import dedent

from sigla import config
from sigla.data.data import Data
from sigla.nodes.abstract_node import AbstractNode
from sigla.templates.engines import TemplateEngineABC
from sigla.nodes.node_list import NodeList
from sigla.utils.helpers import load_module, load_filters_from


class Node(AbstractNode):
    scripts_path = config.path.scripts

    def __init__(
        self,
        tag,
        engine: TemplateEngineABC,
        *,
        attributes=None,
        children=None,
        parent_attributes=None,
        context=None,
    ):

        self.children = children if children is not None else NodeList()
        self.context = context if context is not None else {}

        attributes = attributes if attributes is not None else {}

        parent_attributes = (
            parent_attributes if parent_attributes is not None else {}
        )

        self.data = Data(
            tag=tag,
            attributes=attributes,
            children=self.children,
            parent_attributes=parent_attributes,
        )

        self.engine = engine

        self._handle_script_property()

    def _handle_script_property(self):
        """
        It's possible to pass a script prop to a tag and that script is
        expected to have a "main" function to be run.
        """
        # TODO document this
        if "script" not in self.data.attributes:
            return

        script = self.data.attributes["script"]

        module_path = path.join(self.scripts_path, script)

        new_attrs = _load_and_run_module_main_function(module_path, self)

        # handle returned value
        if new_attrs is None or type(new_attrs) != dict:
            raise ValueError("Expected to receive a dictionary of new args")

        self.data.attributes.update(**new_attrs)

    @property
    def attributes(self):
        return self.data.get_attributes()

    def flatten(self):
        return [self, *self.children.flatten()]

    def append(self, node: AbstractNode):
        self.children.append(node)
        return self

    def render(self, template, **kwargs) -> str:

        node = self
        filters = self.get_filters()

        try:
            return self.engine.render(
                template, filters=filters, **kwargs, node=node
            )
        except Exception as e:
            err_message = _generate_error_message_for_rendering_exception(
                node=self, template=template
            )
            print(err_message)
            raise e

    def finish(self):
        return [child.finish() for child in self.children]

    def process(self):
        self._render_string_attributes()
        #
        context = self._get_attributes_copy()
        for child in self.children:
            child.update_parent_attributes(**context)
        #
        for child in self.children:
            child.process()

    def update_parent_attributes(self, **kwargs):
        self.data.parent_attributes.update(**kwargs)

    def _render_string_attributes(self):
        # update attributes
        attributes = self.data.attributes
        for key, value in attributes.items():
            if type(value) == str:
                attributes[key] = self.render(value, **self.attributes)

    def _get_attributes_copy(self):
        attributes = self.attributes
        context = self.context

        context = context.copy()
        context.update(dict(attributes))

        return context

    @staticmethod
    def get_filters():
        filters = load_filters_from(config.path.filters)
        return filters

    def __getattr__(self, name: str) -> Any:
        if name in self.attributes:
            return self.attributes[name]
        raise AttributeError(f"{self.__class__.__name__}.{name} is invalid.")

    def __repr__(self) -> str:
        tag = self.data.tag
        attrs = self.attributes if len(self.attributes.keys()) > 0 else ""
        children = self.children if len(self.children) > 0 else ""
        return f"""<{tag} {attrs}>{children}</{self.data.tag}>"""

    def __eq__(self, other):
        return self.data == other.data


def _load_and_run_module_main_function(module_path, *args, name="s"):
    # load module and execute it's `main` function
    module = load_module(name, module_path)
    return module.main(*args)


def _generate_error_message_for_rendering_exception(node=None, template=None):
    return dedent(
        f"""\
        ------------------------------
        ERROR WHILE RENDERING TEMPLATE
        ------------------------------

        TEMPLATE:
        {template}

        NODE:
        {node}

        ---

        """
    )



from pathlib import Path

import typer

from sigla import __version__, config
from sigla.cli.actions import Action, NewDefinitionFile, NewFiltersFile
from sigla.main import process
from sigla.utils.errors import TemplateDoesNotExistError
from sigla.utils.helpers import ensure_dirs


class VersionCommand(Action):
    def run(self):
        self.log(f"Version: {__version__}")


class RunCommand(Action):
    def __init__(self, references):
        self.references = references
        self.globs = [
            Path(config.path.definitions).glob(f"{reference}.xml")
            for reference in references
        ]
        self.matches = [match for glob in self.globs for match in glob]

    @staticmethod
    def handle_definition_file_match(match):
        if not match.exists():
            raise typer.Exit(f"✋ The definition(s) do not exists {match}")

        is_dir = match.is_dir()

        if is_dir:
            return

        print(f":: Reading {match}")

        str_xml = match.read_text()
        process("xml", str_xml, factory=None)

    def run(self):

        if len(self.matches) == 0:
            try:
                print(f"✋ No definition(s) found for {self.references}")
            except TemplateDoesNotExistError as e:
                print(e)
                raise typer.Exit(e)

        for match in self.matches:
            self.handle_definition_file_match(match)


class NewCommand(Action):
    def __init__(self, name) -> None:
        super().__init__()
        self.name = name

    def run(self):
        cmd = NewDefinitionFile(config.path.definitions, self.name)
        cmd.run()


class InitCommand(Action):
    def run(self):
        self.log("sigla init")
        self.create_folder(
            config.path.templates,
            config.path.snapshots,
            config.path.definitions,
            config.path.scripts,
        )
        self.create_filters()

    def create_filters(self):
        self.log(f"- checking/creating file {config.path.filters}")
        cmd = NewFiltersFile(
            config.path.root_directory, config.path.filters_filename
        )
        cmd.run()

    def create_folder(self, *args):
        for path in args:
            self.log(f"- creating folder {path}")
            ensure_dirs(path)



import json


def cast_property(name, value):
    casters = {
        "int": int,
        "float": float,
        "json": json.loads,
        "bool": lambda v: v.lower() in ["true", "1"],
    }
    for kind, caster in casters.items():
        dashed_kind = "-" + kind
        if name.endswith(dashed_kind):
            new_key = name.replace(dashed_kind, "")
            new_value = caster(value)
            return new_key, new_value

    return name, value


def cast_dict(attributes: dict) -> dict:
    """
    Takes a dict of props and casts them if necessary
    """
    ret = {}
    for name, value in attributes.items():
        name, value = cast_property(name, value)
        ret[name] = value
    return ret



from typing import Any

from sigla.nodes.node import Node


class NodeEcho(Node):
    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag, *args, **kwargs)
        self.content = None

    def process(self):
        super().process()
        self.content = self.children()
        return self

    def __eq__(self, o: Any) -> bool:
        if type(self) == self.__class__ and type(o) == self.__class__:
            if self.content == o.content:
                return True
        return False

    def finish(self):
        super().finish()
        print(self.content)



from textwrap import dedent

from frontmatter import u, detect_format, handlers
from yaml.parser import ParserError


class FrontMatter:
    @classmethod
    def parse_with_handler(cls, handler, raw_frontmatter, *, metadata=None):
        return cls(handler=handler).parse(raw_frontmatter, metadata=metadata)

    def __init__(self, handler=None):
        self.handler = handler

    def split(self, raw_content, *, encoding="utf-8"):
        raw_content = u(raw_content, encoding).strip()

        # this will only run if a handler hasn't been set higher up
        self.handler = self.handler or detect_format(raw_content, handlers)
        if self.handler is None:
            return None, raw_content, None

        try:
            fm, content = self.handler.split(raw_content)
        except ValueError:
            return None, raw_content, self.handler

        return fm, content, self.handler

    def parse(self, raw_frontmatter, *, metadata=None):
        if metadata is None:
            metadata = {}

        try:
            raw_frontmatter = self.handler.load(raw_frontmatter)
        except ParserError as e:
            print(
                dedent(
                    f"""
            ===
            There is an error on the following yaml (front matter)

            {raw_frontmatter}

            ===

            """
                )
            )
            raise e

        if isinstance(raw_frontmatter, dict):
            metadata.update(raw_frontmatter)

        return metadata

    def get_content(self, template):
        frontmatter, content, handler = self.split(template)
        return content.strip()



from pathlib import Path
from typing import Any

from sigla.nodes.node import Node
from sigla.utils.helpers import ensure_dirs

# TODO
# <models>
#   <model name="User">
#     ...
#   </model>
# </models>
# <forms>
#   <user_form model="User"
#       /or/ model-query="models model[name='User']" /or/ $model="User">
#     ...
#   </user_form>
# </forms>


class NodeFile(Node):
    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag, *args, **kwargs)
        self.content = None

    @property
    def to(self):
        return self.data.attributes["to"]

    def process(self):
        super().process()
        self.content = self.children()
        self.validate_to_attribute()
        return self

    def validate_to_attribute(self):
        # TODO convert this to validation mechanism
        if "to" not in self.data.attributes.keys():
            raise AttributeError(
                "You need to provide the propriety `to` with a filepath on "
                "the element <file> to save the results to"
            )

    def __eq__(self, other: Any) -> bool:
        # TODO apply this logic to other __eq__ methods
        if type(self) == self.__class__ and type(other) == self.__class__:
            if self.content == other.content and self.to == other.to:
                return True
        return False

    def finish(self):
        print(f":: Saving {self.to}")
        ensure_dirs(Path(self.to).parent)
        with open(self.to, "w") as h:
            h.write(self.content)



from abc import ABC, abstractmethod

from sigla.data.data import Data


class AbstractNode(ABC):
    data: Data = None

    # Steps are
    # 1. Load X into Node/Data/Attributes
    # 2. Process
    # 3. Finish

    def __call__(self, *args, **kwargs):
        return self.process()

    @abstractmethod
    def process(self):
        pass

    @abstractmethod
    def finish(self):
        pass
